use core::alloc;
use errors::LoadPEError;
use object::{
    bytes_of, bytes_of_slice, endian,
    pe::IMAGE_DIRECTORY_ENTRY_IMPORT,
    read::pe::{self, ImageOptionalHeader, PeFile32, PeFile64},
    LittleEndian, Object, ObjectSection, Pod,
};
use std::{
    alloc::{alloc, Layout},
    fs,
};
use windows::{
    core::PCSTR,
    Win32::{Foundation::CloseHandle, System::LibraryLoader::LoadLibraryA},
};
pub mod errors;
mod help;

/// 加载并启动一个exe文件，需要支持命令行参数
///
///
///
///
pub fn load_exe(exe_path: &str, args: Option<Vec<&str>>) -> Result<(), LoadPEError> {
    // 1. 检查文件是否存在
    if !fs::metadata(exe_path).is_ok() {
        return Err(LoadPEError::FileNotFound(exe_path.to_string()));
    }
    // 2. 如果文件存在，加载文件
    let data = fs::read(exe_path)?;
    #[cfg(target_arch = "x86")]
    let file: PeFile32 = pe::PeFile::parse(&data[..])?;
    #[cfg(target_arch = "x86_64")]
    let file: PeFile64 = pe::PeFile::parse(&data[..])?;

    // 获取镜像大小
    let image_size = file
        .nt_headers()
        .optional_header
        .size_of_image
        .get(endian::LittleEndian);
    // 3. 为镜像分配内存
    // 分配的内存，后续需要修改内存属性
    //
    let buf = unsafe { alloc(Layout::for_value(&image_size)) };
    if buf.is_null() {
        return Err(LoadPEError::MemoryAllocFailed);
    }
    // 进行数据的拷贝。
    {
        // 拷贝整个头部
        let head_size = file
            .nt_headers()
            .optional_header
            .size_of_headers
            .get(endian::LittleEndian);
        let head_data = &data[..head_size as usize];
        unsafe {
            std::ptr::copy_nonoverlapping(head_data.as_ptr(), buf, head_size as usize);
        }
    }
    {
        // 分块拷贝节区数据

        let section_alignment = file.nt_headers().optional_header.section_alignment();
        let file_alignment = file.nt_headers().optional_header.file_alignment();
        for section in file.sections() {
            // 如果节区没有映射到内存中，那么就不需要拷贝
            if section.pe_section().size_of_raw_data.get(LittleEndian) == 0 {
                continue;
            }
            // VA对齐值
            let va = align_to!(
                section.pe_section().virtual_address.get(LittleEndian),
                section_alignment
            );
            // pointer_to_raw_data对齐值
            let pointer_to_raw_data = align_to!(
                section.pe_section().pointer_to_raw_data.get(LittleEndian),
                file_alignment
            );
            // size_of_raw_data对齐值
            let size_of_raw_data = align_to!(
                section.pe_section().size_of_raw_data.get(LittleEndian),
                file_alignment
            );
            // 从data[pointer_to_raw_data..pointer_to_raw_data + size_of_raw_data]中拷贝数据到buf[va..va + size_of_raw_data]
            let section_data = &data
                [pointer_to_raw_data as usize..(pointer_to_raw_data + size_of_raw_data) as usize];
            unsafe {
                std::ptr::copy_nonoverlapping(
                    section_data.as_ptr(),
                    buf.add(va as usize),
                    size_of_raw_data as usize,
                );
            };
        }
        // 修复IAT表
        {
            let import_table = file.import_table()?;
            if import_table.is_none() {
                return Err(LoadPEError::ExecutableWithoutImportTable);
            }
            let import_table = import_table.unwrap();

            // 1. 加载模块
            let descriptors = import_table.descriptors().unwrap();
            for descriptor_result in descriptors {
                // 当descriptor 是Ok(null)的时候表示结束
                let descriptor = descriptor_result?;
                if descriptor.is_null() {
                    break;
                }
                let module_name_rva = descriptor.name.get(LittleEndian);
                let p_module_name = unsafe {
                    // 从data[module_name_rva..]中读取一个c风格字符串
                    data.as_ptr().add(module_name_rva as usize) as *const u8
                };

                // 使用LoadLibraryA加载模块
                let h_module = unsafe {
                    LoadLibraryA(PCSTR(p_module_name))
                        .map_err(|err| LoadPEError::LoadLibraryError(err))?
                };
                if h_module.is_invalid() {
                    // 这个逻辑应该走不到
                    unimplemented!("LoadLibraryA failed");
                }
                // TODO: 遍历IAT表，修复IAT表
                // 加载每一个方法。使用GetProcAddress，获取地址。
                // 然后修复IAT表。如果其中一个环节出现问题，我们应该：释放所有加载的Module，然后返回错误

                // 如果OriginalFirstThunk不存在，则使用FirstThunk 的值
                

                let original_first_thunk = descriptor.original_first_thunk.get(LittleEndian);
                if original_first_thunk == 0 {

                }else {




                }





            }
        }
    }

    todo!()
}
